# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import time

import pytest
import numpy as np

from qtpy import QtCore
from glue.core import roi
from cubeviz.tools.smoothing import SelectSmoothing

from ...tests.helpers import (toggle_viewer, select_viewer, left_click,
                      left_button_press, right_button_press, enter_slice_text,
                      assert_all_viewer_indices, assert_slice_text)


DATA_LABELS = ['018.DATA', '018.NOISE']


@pytest.fixture(scope='module')
def smoothing_gui(cubeviz_layout):
    cl = cubeviz_layout

    sm = SelectSmoothing(cl._data, parent=cl)

    return sm


def assert_red_stylesheet(widget):
    assert widget.styleSheet() == "color: rgba(255, 0, 0, 128)"

@pytest.mark.parametrize("x", [0,1,2,3,4,5])
def test_smoothing_spatial(qtbot, smoothing_gui, cubeviz_layout, x):
    # TODO: test spectral as well

    # Create GUI
    sm = smoothing_gui
    sm.k_size.setText("1")          # Kernel size
    sm.combo.setCurrentIndex(x)         # Kernel type
    sm.component_combo.setCurrentIndex(0)       # Data componenet

    # Call smoothing
    sm.call_main()
    sm.smooth_cube.thread.wait(2*60*1000)
    qtbot.keyPress(sm.abort_window.info_box, QtCore.Qt.Key_Enter)

    # Get Smoothed data
    smoothed_name = "018.DATA_Smoothed(" + sm.combo.currentText() + ", " + sm.component_combo.currentText()
    smoothing_component_id = [str(x) for x in cubeviz_layout._data.main_components if str(x).startswith(smoothed_name)]
    if isinstance(smoothing_component_id, list) and len(smoothing_component_id) > 0:
        smoothing_component_id = smoothing_component_id[0]
    else:
        return
    np_result = np.asarray(cubeviz_layout._data[smoothing_component_id])

    # Get expected results
    cube = sm.smooth_cube.data_to_cube()
    if "median" == sm.smooth_cube.kernel_type:
        if "spatial" == sm.smooth_cube.smoothing_axis:
            new_cube = cube.spatial_smooth_median(sm.smooth_cube.kernel_size)
        else:
            new_cube = cube.spectral_smooth_median(sm.smooth_cube.kernel_size)
    else:
        kernel = sm.smooth_cube.get_kernel()
        if "spatial" == sm.smooth_cube.smoothing_axis:
            new_cube = cube.spatial_smooth(kernel)
        else:
            new_cube = cube.spectral_smooth(kernel)
    expected_result = new_cube._data

    assert np.allclose(np_result, expected_result, rtol=0.01, equal_nan=True)



import warnings

import numpy as np

from astropy.utils.exceptions import AstropyWarning
from astropy import units as u
from astropy.units.quantity import Quantity
from astropy.wcs.utils import proj_plane_pixel_area

from glue.core.data import ComponentID

from qtpy.QtWidgets import QMessageBox

from .flux_unit_registry import (FLUX_UNIT_REGISTRY, AREA_UNIT_REGISTRY, FORMATTED_UNITS,
                                 NONE_CubeVizUnit, UNKNOWN_CubeVizUnit, ASTROPY_CubeVizUnit,
                                 CUBEVIZ_UNIT_TYPES)

from .flux_units_gui import ConvertFluxUnitGUI


class CubeVizUnit:
    """
    This is unit container for CubeViz. It stores the original
    data unit and a unit representing the current units setting.
    convert_unit function is used to convert astropy units if
    the current unit does not match the original data unit. All
    displays should call the convert_unit function before using the
    data values.
    There are 3 types of CubeVizUnits:
        - NONE_CubeVizUnit: No unit nor unit_string provided (Not Convertible)
        - UNKNOWN_CubeVizUnit: unit_string is provided but no unit (Not Convertible)
        - ASTROPY_CubeVizUnit: Astropy unit provided along with unit_string (Convertible)

    As defined in flux_unit_registry:
        - NONE_CubeVizUnit = "NONE"
        - UNKNOWN_CubeVizUnit = "UNKNOWN"
        - ASTROPY_CubeVizUnit = "ASTROPY"
    """
    def __init__(self, unit=None,
                 unit_string="",
                 component_id=None,
                 unit_type=None):

        # Classify the CubeVizUnit type:
        if unit_type is None:
            if isinstance(unit, u.Unit):
                unit_type = ASTROPY_CubeVizUnit
            elif unit_string:
                unit_type = UNKNOWN_CubeVizUnit
            else:
                unit_type = NONE_CubeVizUnit

        self._controller = None  # Unit controller (property)
        self._original_unit = unit  # the data's actual units
        self._original_unit_string = unit_string  # original_unit as str
        self._unit = unit  # Current Unit
        self._unit_string = unit_string  # unit as str
        self._type = unit_type  # Type of CubeVizUnit
        self.component_id = component_id  # Glue component ID

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit):
        self._unit = unit

    @property
    def unit_string(self):
        return self._unit_string

    @unit_string.setter
    def unit_string(self, unit_string):
        if isinstance(unit_string, str):
            self._unit_string = unit_string

    @property
    def original_unit(self):
        return self._original_unit

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, unit_type):
        if isinstance(type, str):
            self._type = unit_type

    @property
    def controller(self):
        return self._controller

    @controller.setter
    def controller(self, controller):
        self._controller = controller

    def convert_value(self, value, wave=None, new_unit=None):
        if isinstance(value, Quantity):
            value = value.value
        elif not isinstance(value, int) \
                and not isinstance(value, float) \
                and not isinstance(value, np.ndarray):
            raise ValueError("Expected float or int, got {} instead.".format(type(value)))

        if self.type in [NONE_CubeVizUnit, UNKNOWN_CubeVizUnit]:
            return value

        new_value = value

        if new_unit is None:
            new_unit = self._unit

        if hasattr(u.spectral_density, "pixel_area"):
            u.spectral_density.pixel_area = self.controller.pixel_area

        if wave is not None:
            new_value = (value * self._original_unit).to(new_unit, equivalencies=u.spectral_density(wave)).value
        else:
            new_value = (value * self._original_unit).to(new_unit).value

        if isinstance(new_value, Quantity):
            new_value = new_value.value

        return new_value

    def convert_from_original_unit(self, value, wave=None):
        return self.convert_value(value, wave=wave)


class FluxUnitController:
    """
    NOTE: One FluxUnitController per data-set!

    Flux unit controller. Sets up unit containers
    (CubeVizUnit) and stores them. The actual conversion
    happens in the CubeVizUnit and can be done by calling
    the CubeVizUnit.convert_value(...) function.
    Examples
    --------
    To set/add component units:
        FluxUnitController[component_id] = unit
        FluxUnitController.add_component_unit(component_id, unit)
    To set data-set:
        FluxUnitController.set_data(Data)
    To get cubeviz units:
        FluxUnitController[component_id]
        FluxUnitController.get_component_unit(key, cubeviz_unit=True)
    """
    def __init__(self, cubeviz_layout=None):
        self.cubeviz_layout = cubeviz_layout  # CubeViz layout

        self.data = None  # Glue data. Use set_data to add
        self.wcs = None  # WCS info
        self._components = {}  # Dict containing CubeVizUnits

    def __len__(self):
        return len(self.components)

    def __getitem__(self, key):
        return self.get_component_unit(key, cubeviz_unit=True)

    def __setitem__(self, key, value):
        return self.add_component_unit(key, value)

    def __delitem__(self, key):
        return self.remove_component_unit(key)

    def __iter__(self):
        return self.components.__iter__()

    @property
    def components(self):
        return self._components

    @property
    def pixel_area(self):
        if self.wcs is None:
            return None
        try:
            top_unit = u.Unit(self.wcs.wcs.cunit[0])
            pixel_area = (proj_plane_pixel_area(self.wcs) * (top_unit ** 2) / u.pix).to(u.arcsec ** 2 / u.pix)
            return pixel_area
        except (ValueError, AttributeError):
            return None

    @property
    def wave(self):
        if self.cubeviz_layout:
            return self.cubeviz_layout.get_wavelength()
        return None

    @property
    def wavelengths(self):
        if self.cubeviz_layout:
            waves = self.cubeviz_layout.get_wavelengths()
            units = self.cubeviz_layout.get_wavelengths_units()
            return waves * units
        return None

    def construct_3d_wavelengths(self, component):
        if self.cubeviz_layout:
            waves = self.cubeviz_layout.get_wavelengths()
            units = self.cubeviz_layout.get_wavelengths_units()
            waves3d = np.zeros_like(component)
            for i in range(len(waves3d)):
                waves3d[i] = waves[i]
            return waves3d * units
        return None

    @staticmethod
    def string_to_unit(unit_string):
        """
        Unit str -> astropy unit
        :param unit_string: str
        :return: astropy unit. None otherwise
        """
        try:
            if unit_string:
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', AstropyWarning)
                    astropy_unit = u.Unit(unit_string)
                return astropy_unit
        except ValueError:
            print("Warning: Failed to convert {0} to an astropy unit.".format(unit_string))
        return None

    @staticmethod
    def unit_to_string(unit):
        """
        astropy unit -> unit str
        :param unit: astropy unit
        :return: str
        """
        if unit is None:
            return ""
        elif isinstance(unit, str):
            return unit
        elif isinstance(unit, u.UnitBase):
            return unit.to_string()
        elif isinstance(unit, Quantity):
            return unit.unit.to_string()
        else:
            raise ValueError("Invalid input unit type {0}.".format(type(unit)))

    def add_component_unit(self, component_id, unit=None):
        """
        Add or update componet unit.
        :param component_id: component id or str
        :param unit: string or astropy unit
        :return: CubeVizUnit
        """

        if not isinstance(component_id, ComponentID):
            raise Exception('component_id must be of type ComponentID, but was {}'.format(type(component_id)))

        if isinstance(unit, CubeVizUnit):
            unit.controller = self
            self._components[component_id] = unit
            return unit

        unit_string = self.unit_to_string(unit)

        # IF: empty unit or no unit
        if not unit_string:
            astropy_unit = None
            unit_string = ""
            unit_type = NONE_CubeVizUnit
        # ELSE IF: formatted unit
        elif unit_string in FORMATTED_UNITS:
            registered_unit = FORMATTED_UNITS[unit_string]
            if "astropy_unit_string" in registered_unit:
                unit_string = registered_unit["astropy_unit_string"]
            astropy_unit = self.string_to_unit(unit_string)
            unit_type = ASTROPY_CubeVizUnit
        else:
            astropy_unit = self.string_to_unit(unit_string)
            # IF: no astropy unit component
            if astropy_unit is None:
                unit_type = UNKNOWN_CubeVizUnit
            else:
                unit_type = ASTROPY_CubeVizUnit

        cubeviz_unit = CubeVizUnit(unit=astropy_unit,
                                   unit_string=unit_string,
                                   component_id=component_id,
                                   unit_type=unit_type)

        cubeviz_unit.controller = self
        self._components[component_id] = cubeviz_unit
        return cubeviz_unit

    def remove_component_unit(self, component_id):
        """
        Remove component from controller
        :param component_id: component id
        """
        if component_id in self._components:
            del self._components[component_id]

    def get_component_unit(self, component_id, cubeviz_unit=False):
        """
        Get component units
        :param component_id: component id
        :param cubeviz_unit: if True: return CubeVizUnit
                             else: return astropy unit
        :return: CubeVizUnit, astropy unit or None
        """
        if component_id in self._components:
            if cubeviz_unit:
                return self._components[component_id]
            else:
                return self._components[component_id].unit
        return None

    def set_data(self, data):
        """
        Add glue data and its components to controller.
        glue components' units attribute should contain a
        string with desired units for that component. This
        controller will use the string in that attribute
        to construct and assign units.
        :param data: glue data
        :return:
        """
        self.data = data
        self._components = {}
        for comp in data.visible_components:
            self.add_component_unit(comp, data.get_component(comp).units)

        wcs = data.coords.wcs
        if wcs is not None:
            self.wcs = wcs

    def converter(self, parent=None, convert_data=False):
        """
        Launch Converter GUI
        :param parent: application
        :param values:
            convert data values if True.
            Only displayed units if False.
        :return: ConvertFluxUnitGUI instance
        :raise: Exception: If wavelength info is missing
        """
        if self.wavelengths is None:
            message = "Could not launch conversion ui because" \
                      "wavelength information is missing."
            info = QMessageBox.critical(parent, "Error", message)
            raise Exception(message)

        if convert_data:
            message = "You are about to launch the data flux units converter. " \
                      "This will modify the actual values stored in components " \
                      "according to the unit conversions selected. If you would rather" \
                      " change the displayed units, please select the Convert" \
                      " Displayed Units option."
            info = QMessageBox.warning(parent, "Info", message)

        if hasattr(u.spectral_density, "pixel_area"):
            u.spectral_density.pixel_area = self.pixel_area

        ex = ConvertFluxUnitGUI(self, parent, convert_data)
        return ex



import numpy as np

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QDialog,  QPushButton, QLabel, QHBoxLayout, QVBoxLayout,
    QComboBox, QLineEdit, QWidgetItem, QSpacerItem, QMessageBox
)

from glue.utils.qt import update_combobox

from astropy import units as u
from astropy.units.quantity import Quantity

from ..messages import FluxUnitsUpdateMessage

from .flux_unit_registry import (FLUX_UNIT_REGISTRY, AREA_UNIT_REGISTRY, FORMATTED_UNITS,
                                 NONE_CubeVizUnit, UNKNOWN_CubeVizUnit, ASTROPY_CubeVizUnit,
                                 CUBEVIZ_UNIT_TYPES)


def find_unit_index(unit_list, target_unit):
    """
    Given a list of units or strings, find
    index of unit.
    :param unit_list: list of units or string
    :param target_unit: unit or string
    :return: unit index
    """

    # Make both a string and astropy version
    if not isinstance(target_unit, str):
        target_unit = u.Unit(target_unit).to_string()

    # Create list of all astropy Unit things
    temp_list = [unit.to_string() if not isinstance(unit, str)
                 else u.Unit(unit).to_string() for unit in unit_list]

    if target_unit in temp_list:
        return temp_list.index(target_unit)
    return None


def _get_power(numeric):
    """
    Given a number, get the power if
    the power is an int. Else Return 0
    :param numeric: float or int
    :return: power if int, 0.0 if power is float
    """
    exponent = float(np.log10(numeric))
    return int(exponent) if exponent.is_integer() else 0


class CubeVizUnitLayout:
    """
    This implements a specialized horizontal layout
    in the converter gui. This layout contains
    inputs specific to the unit type. This class
    is responsible for the following tasks:
        - Populate the UI horizontal layout
        - Contain user input widgets (QLineEdit etc..)
        - Set displayed messages on the UI
        - Get units from the user input
    """
    def __init__(self, cubeviz_unit):
        self.cubeviz_unit = cubeviz_unit

        self._unit = cubeviz_unit.unit
        self._unit_string = cubeviz_unit.unit_string

        self.message_box = None  # Pointer to gui message box

    def set_message_box(self, message_box):
        """
        Sets local attribute to message box on
        the onversion gui.
        :param message_box: QLabel
        """
        self.message_box = message_box
        self.message_box.setText("")

    def change_units(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """
        return True

    def reset_widgets(self):
        """
        Reset all attributes pointing to widgets in the
        conversion gui
        """
        self.message_box = None
        return

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets.
        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """

        if self.cubeviz_unit.unit_string:
            default_message = "CubeViz can not convert this unit: {0}."
            default_message = default_message.format(self.cubeviz_unit.unit_string)
        else:
            default_message = "No Units."

        default_label = QLabel(default_message)
        unit_layout.addWidget(default_label)
        return unit_layout

    @property
    def new_unit(self):
        return self._unit


class AstropyUnitLayout(CubeVizUnitLayout):
    """
    This implements a specialized horizontal layout
    in the converter gui. This layout contains
    inputs specific to the unit type. This class
    sets up the layout for regular astropy units.
    """
    def __init__(self, cubeviz_unit):
        super(AstropyUnitLayout, self).__init__(cubeviz_unit)

        self.options_combo = None

    def reset_widgets(self):
        self.message_box = None
        self.options_combo = None

    def change_units(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """

        new_unit_string = self.options_combo.currentText()

        self._unit_string = new_unit_string
        self._unit = u.Unit(new_unit_string)
        return True

    def _update_message(self):
        """
        Callback for when options in conversion gui change. Message
        is set to info about the units a preview of the conversion.
        """
        if self.message_box is None:
            return

        new_unit_string = self.options_combo.currentText()
        new_unit = u.Unit(new_unit_string)

        new_value = self.cubeviz_unit.convert_value(1.0, new_unit=new_unit)

        message_param = (self.cubeviz_unit.original_unit.to_string(), new_unit_string, new_value)

        if 0.01 <= abs(new_value) <= 1000:
            message = "Original Units: [{0}]\n" \
                      "New Units: [{1}]\n" \
                      "1.0 [Original Units] = {2:.2f} [New Units]".format(*message_param)
        else:
            message = "Original Units: [{0}]\n" \
                      "New Units: [{1}]\n" \
                      "1.0 [Original Units] = {2:0.2e} [New Units]".format(*message_param)
        self.message_box.setText(message)

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets.

        Layouts:
            [QComboBox]

        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """

        unit_str = self._unit.to_string()
        options = self._unit.find_equivalent_units(include_prefix_units=True)
        options = [i.to_string() for i in options]
        options.sort(key=lambda x: x.upper())
        if unit_str not in options:
            options.append(unit_str)
        index = options.index(unit_str)
        combo = QComboBox()
        # combo.setFixedWidth(200)
        combo.addItems(options)
        combo.setCurrentIndex(index)
        combo.currentIndexChanged.connect(self._update_message)
        self.options_combo = combo
        unit_layout.addWidget(combo)
        self._update_message()
        return unit_layout


class SpectralFluxDensityLayout(CubeVizUnitLayout):
    """
    This implements a specialized horizontal layout
    in the converter gui. This layout contains
    inputs specific to the unit type. This class
    sets up the layout for regular astropy
    spectral flux density units.
    """
    def __init__(self,
                 cubeviz_unit,
                 power=None,
                 spectral_flux_density=None,
                 area=None,
                 wave=None,
                 pixel_area=None):

        super(SpectralFluxDensityLayout, self).__init__(cubeviz_unit)

        self.power = power
        self.spectral_flux_density = spectral_flux_density
        self.area = area

        self._original_unit = cubeviz_unit.original_unit

        self.has_area = True if area is not None else False

        self.power_input = None
        self.flux_combo = None
        self.area_combo = None

        if wave is not None:
            self.wave = wave
        else:
            self.wave = 656.3 * u.nm  # Used for conversion preview. Updated latter.

        self.pixel_area = pixel_area

    def reset_widgets(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """
        self.message_box = None
        self.power_input = None
        self.flux_combo = None
        self.area_combo = None

    def _get_current_power(self):
        """
        :return: current power, 0 if none
        """
        if self.power is not None:
            power = self.power
        else:
            power = 0
        return power

    def _validate_input(self):
        """
        Validate user input from converter gui.
        :return: bool: True if input is valid
        """
        red = "background-color: rgba(255, 0, 0, 128);"
        success = True
        if self.power_input is None \
                or self.flux_combo is None:
            success = False
            return success

        if self.area_combo is None and self.has_area:
            success = False
            return success

        # Check 1: power_input
        if self.power_input == "":
            self.power_input.setStyleSheet(red)
            success = False
        else:
            try:
                power = int(self.power_input.text())
                self.power_input.setStyleSheet("")
            except ValueError:
                self.power_input.setStyleSheet(red)
                success = False
        return success

    def change_units(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """

        success = self._validate_input()
        if not success:
            return success

        self.power = int(self.power_input.text())

        flux_string = self.flux_combo.currentText()
        self.spectral_flux_density = u.Unit(flux_string)

        if self.has_area:
            area_string = self.area_combo.currentText()
            area_unit = u.Unit(area_string)
            self.area = area_unit
            unit_base = self.spectral_flux_density / self.area
        else:
            unit_base = self.spectral_flux_density

        self._unit = u.Unit((10**self.power) * unit_base)
        self._unit_string = self._unit.to_string()

        return success

    def _update_message(self):
        """
        Callback for when options in conversion gui change. Message
        is set to info about the units a preview of the conversion.
        """
        if self.message_box is None:
            return
        success = self._validate_input()
        if not success:
            self.message_box.setText("Error")
            return

        power = int(self.power_input.text())

        flux_string = self.flux_combo.currentText()
        spectral_flux_density = u.Unit(flux_string)

        if self.has_area:
            area_string = self.area_combo.currentText()
            area = u.Unit(area_string)
            unit_base = spectral_flux_density / area
        else:
            unit_base = spectral_flux_density

        unit_base = u.Unit((10**power) * unit_base)

        if isinstance(unit_base, Quantity):
            unit_base = u.Unit(unit_base)

        new_value = self.cubeviz_unit.convert_value(1.0,
                                                    wave=self.wave,
                                                    new_unit=unit_base)

        if self.pixel_area is not None:
            pixel_area = "{0:.2e}".format(self.pixel_area)
        else:
            pixel_area = "N/A"

        message_param = (self.cubeviz_unit.original_unit.to_string(),
                         unit_base,
                         pixel_area,
                         new_value,
                         self.wave)
        if 0.001 <= abs(new_value) <= 1000:
            message = "Original Units: [{0}]\n\n" \
                      "New Units: [{1}]\n\n" \
                      "Pixel Area (Scale): {2}\n\n" \
                      "1.000 [Original Units] = {3:.3f} [New Units]\n" \
                      "(at lambda = {4:0.4e})".format(*message_param)
        else:
            message = "Original Units: [{0}]\n\n" \
                      "New Units: [{1}]\n\n" \
                      "Pixel Area (Scale): {2}\n\n" \
                      "1.0 [Original Units] = {3:0.2e} [New Units]\n" \
                      "(at lambda = {4:0.4e})".format(*message_param)

        self.message_box.setText(message)

    def _on_flux_combo_change(self, index):
        """
        Callback for flux combo
        :param index:
        :return:
        """
        current_string = self.flux_combo.currentText()
        flux_unit_str = self.spectral_flux_density.to_string()
        if current_string != flux_unit_str:
            self.power_input.setText("0")
        else:
            power = self._get_current_power()
            self.power_input.setText(str(power))

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets. Set wave attribute to current
        wavelength.

        Layouts:
            10^[QLineEdit] X [QComboBox] / [QComboBox]
                        or
            10^[QLineEdit] X [QComboBox]

        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """

        power = self._get_current_power()

        unit_layout.addWidget(QLabel("10^"))
        power_input = QLineEdit(str(power))
        power_input.setFixedWidth(30)
        self.power_input = power_input
        power_input.textChanged.connect(self._update_message)
        unit_layout.addWidget(power_input)

        unit_layout.addWidget(QLabel("   X   "))

        flux_unit_str = self.spectral_flux_density.to_string()
        flux_options = FLUX_UNIT_REGISTRY.compose_unit_list(current_unit=flux_unit_str)
        if flux_unit_str in flux_options:
            index = flux_options.index(flux_unit_str)
        else:
            index = find_unit_index(flux_options, flux_unit_str)
            if index is None:
                flux_options.append(flux_unit_str)
                index = flux_options.index(flux_unit_str)
        flux_combo = QComboBox()
        flux_combo.addItems(flux_options)
        flux_combo.setCurrentIndex(index)
        flux_combo.currentIndexChanged.connect(self._on_flux_combo_change)
        flux_combo.currentIndexChanged.connect(self._update_message)
        self.flux_combo = flux_combo
        unit_layout.addWidget(flux_combo)

        if self.area is not None:
            division = QLabel("   /   ")
            unit_layout.addWidget(division)

            area_str = self.area.to_string()

            no_pixel_area = self.pixel_area is None

            if self.area.decompose() == u.pix.decompose() and no_pixel_area:
                area_options = AREA_UNIT_REGISTRY.compose_unit_list(pixel_only=True)
            elif 'solid angle' in self.area.physical_type and no_pixel_area:
                area_options = AREA_UNIT_REGISTRY.compose_unit_list(solid_angle_only=True)
            else:
                area_options = AREA_UNIT_REGISTRY.compose_unit_list()

            if area_str in area_options:
                index = area_options.index(area_str)
            else:
                index = find_unit_index(area_options, area_str)
                if index is None:
                    area_options.append(area_str)
                    index = area_options.index(area_str)
            area_combo = QComboBox()
            area_combo.width()
            area_combo.addItems(area_options)
            area_combo.setCurrentIndex(index)
            area_combo.currentIndexChanged.connect(self._update_message)
            self.area_combo = area_combo
            unit_layout.addWidget(area_combo)
        unit_layout.addStretch(1)
        unit_layout.setSpacing(0)

        if self.message_box is not None:
            self._update_message()

        return unit_layout


def decompose_sfd_over_solid_angle(unit):
    """
    Decompose sfd_over_solid_angle to CubeVizUnit
    NB: SFD_over_solid_angle = spectral_flux_density / (degree^2)
    :param unit: astropy unit
    :param unit_string: astropy unit
    :return: CubeVizUnit
    """
    power = _get_power(unit.scale)
    unit_list = unit.bases
    power_list = unit.powers

    index = None
    sfd_unit = None
    for i, un in enumerate(unit_list):
        if 'solid angle' == un.physical_type or \
                'angle' == un.physical_type:
            index = i
        else:
            if sfd_unit is None:
                sfd_unit = un ** power_list[i]
            else:
                sfd_unit *= un ** power_list[i]

    if index is not None:
        angle_unit = unit_list[index] ** abs(power_list[index])
        if power == 0:
            sfd_unit = u.Unit(unit.scale * sfd_unit)

        return [power, sfd_unit, angle_unit]
    else:
        return [None, unit, None]


def decompose_sfd_over_pix(unit):
    """
    Decompose sfd_over_pix to CubeVizUnit
    NB: SFD_over_pix = spectral_flux_density / pixel
    :param unit: astropy unit
    :param unit_string: astropy unit
    :return: CubeVizUnit
    """
    power = _get_power(unit.scale)
    unit_list = unit.bases
    power_list = unit.powers

    index = None
    sfd_unit = None
    for i, un in enumerate(unit_list):
        if un.decompose() == u.pix.decompose():
            index = i
        else:
            if sfd_unit is None:
                sfd_unit = un ** power_list[i]
            else:
                sfd_unit *= un ** power_list[i]

    if index is not None:
        pix_unit = unit_list[index] ** abs(power_list[index])
        if power == 0:
            sfd_unit = u.Unit(unit.scale * sfd_unit)
        return [power, sfd_unit, pix_unit]
    else:
        return [None, unit, None]


def assign_cubeviz_unit_layout(cubeviz_unit, pixel_area=None, wave=None):
    """
    This is used to break down the unit and assign the
    appropriate conversion layout.
    Potential unit types are:
        - no units or not-astropy units -> CubeVizUnitLayout
        - regular astropy unit -> AstropyUnitLayout
        - spectral_flux_density -> SpectralFluxDensityLayout
        - spectral_flux_density / solid_angle -> SpectralFluxDensityLayout
        - spectral_flux_density / pixel -> SpectralFluxDensityLayout
    :param cubeviz_unit: CubeVizUnit
    :param pixel_area: Pixel area to use for conversion previews
    :param wave: Wavelength used for conversion previews
    :return:
    """

    if cubeviz_unit.type in [NONE_CubeVizUnit, UNKNOWN_CubeVizUnit]:
        layout = CubeVizUnitLayout(cubeviz_unit)
    else:
        astropy_unit = cubeviz_unit.unit
        # IF: astropy unit type is spectral_flux_density / solid_angle
        if astropy_unit.is_equivalent((u.Jy / u.degree ** 2),
                                      equivalencies=u.spectral_density.get_basic_relations(656.3*u.nm)):
            power, sfd_unit, area = decompose_sfd_over_solid_angle(astropy_unit)
            layout = SpectralFluxDensityLayout(cubeviz_unit, power, sfd_unit, area,
                                               wave, pixel_area)
        # ELSE IF: astropy unit type is spectral_flux_density / pixel
        elif astropy_unit.is_equivalent((u.Jy / u.pix),
                                        equivalencies=u.spectral_density.get_basic_relations(656.3*u.nm)):
            power, sfd_unit, area = decompose_sfd_over_pix(astropy_unit)
            layout = SpectralFluxDensityLayout(cubeviz_unit, power, sfd_unit, area,
                                               wave, pixel_area)
        # ELSE IF: astropy unit type is spectral_flux_density
        elif 'spectral flux density' in astropy_unit.physical_type:
            sfd_unit = astropy_unit
            numeric = sfd_unit.scale
            power = _get_power(numeric)
            layout = SpectralFluxDensityLayout(cubeviz_unit, power, sfd_unit, None,
                                               wave, pixel_area)
        # ELSE: astropy unit type is not special
        else:
            layout = AstropyUnitLayout(cubeviz_unit)

    return layout


class ConvertFluxUnitGUI(QDialog):
    """
    GUI for unit conversions
    """
    def __init__(self, controller, parent=None, convert_data=False):
        super(ConvertFluxUnitGUI, self).__init__(parent=parent)
        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        self.title = "Unit Conversion"
        self.setMinimumSize(400, 270)

        self.convert_data = convert_data

        self.cubeviz_layout = controller.cubeviz_layout
        self._hub = self.cubeviz_layout.session.hub

        self.controller = controller
        self.data = controller.data
        self.controller_components = controller._components

        self.current_unit = None
        self.current_layout = None

        self._init_ui()

    def _init_ui(self):
        # LINE 1: Data component drop down
        self.component_prompt = QLabel("Data Component:")
        self.component_prompt.setWordWrap(True)
        # Add the data component labels to the drop down, with the ComponentID
        # set as the userData:
        if self.parent is not None and hasattr(self.parent, 'data_components'):
            self.label_data = [(str(cid), cid) for cid in self.parent.data_components]
        else:
            self.label_data = [(str(cid), cid) for cid in self.data.visible_components]

        default_index = 0
        self.component_combo = QComboBox()
        self.component_combo.setFixedWidth(200)
        update_combobox(self.component_combo, self.label_data, default_index=default_index)
        self.component_combo.currentIndexChanged.connect(self.update_unit_layout)

        # hbl is short for Horizontal Box Layout
        hbl1 = QHBoxLayout()
        hbl1.addWidget(self.component_prompt)
        hbl1.addWidget(self.component_combo)
        hbl1.addStretch(1)

        # LINE 2: Unit conversion layout
        # This layout is filled by CubeVizUnit
        self.unit_layout = QHBoxLayout()  # this is hbl2

        # LINE 3: Message box
        self.message_box = QLabel("")
        hbl3 = QHBoxLayout()
        hbl3.addWidget(self.message_box)
        hbl3.addStretch(1)

        # Line 4: Buttons
        ok_text = "Convert Data" if self.convert_data else "Convert Displayed Units"
        ok_function = self.convert_data_units if self.convert_data else self.convert_displayed_units
        self.okButton = QPushButton(ok_text)
        self.okButton.clicked.connect(ok_function)
        self.okButton.setDefault(True)

        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.cancel)

        hbl4 = QHBoxLayout()
        hbl4.addStretch(1)
        hbl4.addWidget(self.cancelButton)
        hbl4.addWidget(self.okButton)

        vbl = QVBoxLayout()
        vbl.addLayout(hbl1)
        vbl.addLayout(self.unit_layout)
        vbl.addLayout(hbl3)
        vbl.addLayout(hbl4)
        self.setLayout(vbl)
        self.vbl = vbl

        self.update_unit_layout(default_index)

        self.show()

    def update_unit_layout(self, index):
        """
        Call back for component selection drop down.
        """
        component_id = self.component_combo.currentData()

        # STEP1: Clean up widgets from last component
        widgets = (self.unit_layout.itemAt(i) for i in range(self.unit_layout.count()))
        for w in widgets:
            if isinstance(w, QSpacerItem):
                self.unit_layout.removeItem(w)
                continue
            elif isinstance(w, QWidgetItem):
                w = w.widget()

            if hasattr(w, "deleteLater"):
                w.deleteLater()

        self.message_box.setText("")

        if self.current_layout:
            self.current_layout.reset_widgets()

        # STEP2: Add now component and connect to CubeVizUnit
        #        so that new widgets are populated.
        if component_id in self.controller_components:
            cubeviz_unit = self.controller_components[component_id]
            self.current_unit = cubeviz_unit

            wave = self.controller.wave
            pixel_area = self.controller.pixel_area
            layout = assign_cubeviz_unit_layout(cubeviz_unit,
                                                wave=wave,
                                                pixel_area=pixel_area)
            layout.set_message_box(self.message_box)
            layout.populate_unit_layout(self.unit_layout, self)
            self.current_layout = layout
            if ASTROPY_CubeVizUnit == cubeviz_unit.type:
                self.okButton.setEnabled(True)
            else:
                self.okButton.setEnabled(False)
        else:
            self.current_unit = None
            default_message = "CubeViz can not convert this unit."
            default_label = QLabel(default_message)
            self.unit_layout.addWidget(default_label)
            self.okButton.setEnabled(False)

        self.unit_layout.update()
        self.vbl.update()

    def convert_displayed_units(self):
        """
        Calls CubeVizUnit.change_units to finalize
        conversions. Updates plots with new units.
        :return:
        """
        success = self.current_layout.change_units()
        if not success:
            info = QMessageBox.critical(self, "Error", "Conversion failed.")
            return

        new_unit = self.current_layout.new_unit
        self.current_unit.unit = new_unit
        self.current_unit.unit_string = str(new_unit)

        component_id = self.component_combo.currentData()
        self.data.get_component(component_id).units = self.current_unit.unit_string
        msg = FluxUnitsUpdateMessage(self, self.current_unit, component_id)
        self._hub.broadcast(msg)
        self.close()

    def convert_data_units(self):
        """
        Calls CubeVizUnit.change_units to finalize
        conversions. Updates plots with new units.
        :return:
        """
        success = self.current_layout.change_units()
        if not success:
            info = QMessageBox.critical(self, "Error", "Conversion failed.")
            return

        new_unit = self.current_layout.new_unit
        self.current_unit.unit = new_unit
        self.current_unit.unit_string = str(new_unit)

        component_id = self.component_combo.currentData()
        component = component_id.parent.get_component(component_id)

        old_array = component._data.copy()
        old_array.flags.writeable = True

        wavelengths = self.controller.construct_3d_wavelengths(old_array)

        new_array = self.current_unit.convert_value(old_array, wave=wavelengths)

        component._data = new_array

        self.current_unit = self.controller.add_component_unit(component_id, new_unit)
        component.units = self.current_unit.unit_string
        msg = FluxUnitsUpdateMessage(self, self.current_unit, component_id)
        self._hub.broadcast(msg)
        self.close()

    def cancel(self):
        self.close()



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure.

import os
import sys

from astropy.version import version as astropy_version
if astropy_version < '3.0':
    # With older versions of Astropy, we actually need to import the pytest
    # plugins themselves in order to make them discoverable by pytest.
    from astropy.tests.pytest_plugins import *
else:
    # As of Astropy 3.0, the pytest plugins provided by Astropy are
    # automatically made available when Astropy is installed. This means it's
    # not necessary to import them here, but we still need to import global
    # variables that are used for configuration.
    from astropy.tests.plugins.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

from astropy.tests.helper import enable_deprecations_as_exceptions

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
## as follow (although default should work for most cases).
## To ignore some packages that produce deprecation warnings on import
## (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
## 'setuptools'), add:
##     modules_to_ignore_on_import=['module_1', 'module_2']
## To ignore some specific deprecation warning messages for Python version
## MAJOR.MINOR or later, add:
##     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# enable_deprecations_as_exceptions()

## Uncomment and customize the following lines to add/remove entries from
## the list of packages for which version numbers are displayed when running
## the tests. Making it pass for KeyError is essential in some cases when
## the package uses other astropy affiliated packages.
try:
    PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
    PYTEST_HEADER_MODULES['Glue'] = 'glue'
    PYTEST_HEADER_MODULES['Asdf'] = 'asdf'
    PYTEST_HEADER_MODULES['QtPy'] = 'qtpy'
    PYTEST_HEADER_MODULES['SpecViz'] = 'specviz'
except (NameError, KeyError):  # NameError is needed to support Astropy < 1.0
    pass

## Uncomment the following lines to display the version number of the
## package rather than the version number of Astropy in the top line when
## running the tests.
import os

# This is to figure out the package version, rather than
# using Astropy's
try:
    from . import __version__ as version
except ImportError:
    version = 'dev'

try:
    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version
except NameError:   # Needed to support Astropy <= 1.0.0
    pass


import pytest
from .tests.helpers import (toggle_viewer, select_viewer, create_glue_app,
                            reset_app_state)

if not os.environ.get('JWST_DATA_TEST', False):
    @pytest.fixture(scope='session')
    def cubeviz_layout():
        app = create_glue_app()
        layout = app.tab(0)

        # Cheap workaround for Windows test environment
        if sys.platform.startswith('win'):
            layout._cubeviz_toolbar._toggle_sidebar()

        return app.tab(0)

    @pytest.fixture(autouse=True)
    def reset_state(qtbot, cubeviz_layout):
        # This yields the test itself
        yield

        # Make sure to return the application to this state between tests
        reset_app_state(qtbot, cubeviz_layout)



import sys
import traceback
import multiprocessing
from urllib.parse import urljoin

import pytest

from specviz.tests.test_load_data import (BOX_PREFIX, jwst_data_test,
                                          download_test_data,
                                          run_subprocess_test)

from cubeviz.cubeviz import create_app


JWST_DATA_FILES = [
    # det_image_ch1-short_s3d.fits
    'mhicn6ofnwpi4mzjm4yoppm95mdhmw42.fits',
    # det_image_ch1-short_s3d_ref.fits
    'k4kwyn1t5zcjw7wg4xv9ub3fgg07lwl7.fits',
    # det_image_ch2-short_s3d.fits
    'kpye05xrwcpllzlju0l96nw0lmtwom98.fits',
    # ditherunity_CLEAR_PRISM_M1_m1_noiseless_NRS1_modified_updatedHDR_fixintarget_s3d.fits
    'f3bozjgdgtv0n1fqdt002ufmf1z97qup.fits',
]

JWST_DATA_PATHS = [urljoin(BOX_PREFIX, name) for name in JWST_DATA_FILES]


def run_cubeviz_test(q, _, tmpdir, url, *args):
    app = None

    try:
        fname = download_test_data(tmpdir, url)
        app = create_app(interactive=False)
        app.load_data([fname])
        assert len(app.data_collection) == 1
        assert app.data_collection[0].label.startswith('jwst-fits-cube')
    except Exception:
        ex_type, ex_value, tb = sys.exc_info()
        error = ex_type, ex_value, ''.join(traceback.format_tb(tb))
    else:
        error = None
    finally:
        if app is not None:
            app.app.quit()

    q.put(error)


@jwst_data_test
@pytest.mark.parametrize('url', JWST_DATA_PATHS)
def test_load_jwst_data(tmpdir, url):
    multiprocessing.set_start_method('spawn', force=True)
    run_subprocess_test(None, tmpdir, url, callback=run_cubeviz_test)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This is an Astropy affiliated package.
"""

__minimum_python_version__ = "3.6"

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._internal_init import *
# ----------------------------------------------------------------------------

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = "unknown"


# This *must* occur before any imports from glue. It prevents the "python must
# be installed as a framework" error that occurs on OSX. Since we now
# explicitly depend on PyQt5, we use it as the mpl backend.
import matplotlib as mpl
mpl.use('Qt5Agg')

import astropy.units as units
from .flux_equivalences import CustomFluxEquivalences
# We override the units.equivalencies.spectral_density function with
# CustomFluxEquivalences before the program starts. We expect all libraries
# to access CustomFluxEquivalences when calling for units.equivalencies.spectral_density
units.equivalencies.spectral_density = CustomFluxEquivalences(units.equivalencies.spectral_density)
units.spectral_density = units.equivalencies.spectral_density

from . import keyboard_shortcuts



import pytest
import numpy as np
from glue.core import roi

from cubeviz.utils.contour import ContourSettings
from cubeviz.tools.moment_maps import MomentMapsGUI


@pytest.fixture(scope='module')
def moment_maps_gui(cubeviz_layout):
    cl = cubeviz_layout

    mm = MomentMapsGUI(cl._data, cl.session.data_collection, parent=cl)

    return mm


def test_stats_box_without_subset(cubeviz_layout):
    """
    Tests the stat box underneath the ImageViewer when it is the full spectrum
    """
    cl_viewer = cubeviz_layout.split_views[1]._widget

    cl_viewer._subset = None

    data = cl_viewer._data[0][cl_viewer.current_component_id][cl_viewer._slice_index]

    wave = cl_viewer.cubeviz_layout.get_wavelength(cl_viewer.slice_index)
    data_wave = cl_viewer.cubeviz_unit.convert_value(data, wave=wave)

    assert data_wave is not None

    results = (np.nanmin(data_wave), np.nanmax(data_wave), np.median(data_wave), data_wave.mean(), data_wave.std())
    results_string = r"min={:.4}, max={:.4}, median={:.4}, μ={:.4}, σ={:.4}".format(*results)

    assert results_string == cl_viewer.parent().stats_text.text()


def test_stats_box_with_subset(cubeviz_layout):
    """
    Tests the stat box underneath the ImageViewer when there is an ROI
    """
    cl_viewer = cubeviz_layout.split_views[1]._widget

    # Create a subset (ROI) if there is none
    cl_viewer.apply_roi(roi.CircularROI(xc=6, yc=10, radius=3))

    assert cl_viewer._subset is not None

    mask = cl_viewer._subset.to_mask()[cl_viewer._slice_index]
    data = cl_viewer._data[0][cl_viewer.current_component_id][cl_viewer._slice_index][mask]

    wave = cl_viewer.cubeviz_layout.get_wavelength(cl_viewer.slice_index)
    data_wave = cl_viewer.cubeviz_unit.convert_value(data, wave=wave)

    assert data_wave is not None

    results = (np.nanmin(data_wave), np.nanmax(data_wave), np.median(data_wave), data_wave.mean(), data_wave.std())
    results_string = r"min={:.4}, max={:.4}, median={:.4}, μ={:.4}, σ={:.4}".format(*results)

    assert results_string == cl_viewer.parent().stats_text.text()

    # Remove the ROI/subset
    dc = cubeviz_layout.session.application.data_collection
    dc.remove_subset_group(dc.subset_groups[0])


def test_overlay(moment_maps_gui, cubeviz_layout):
    # Only "No Overlay" option available in combobox
    assert cubeviz_layout._overlay_controller._overlay_image_combo.count() == 1

    # Create moment map GUI
    mm = moment_maps_gui
    mm.display()
    mm.order_combobox.setCurrentIndex(0)
    mm.data_combobox.setCurrentIndex(0)

    # Call calculate function and get result
    mm.calculateButton.click()

    # Second option in combobox for moment map overlay
    assert cubeviz_layout._overlay_controller._overlay_image_combo.count() == 2

    cl_viewer = cubeviz_layout.split_views[1]._widget
    assert len(cl_viewer.axes.images) == 1

    # Set overlay and check changing of settings works
    cubeviz_layout._overlay_controller._overlay_image_combo.setCurrentIndex(1)
    cubeviz_layout._overlay_controller._overlay_colormap_combo.setCurrentIndex(10)
    cubeviz_layout._overlay_controller._alpha_slider.setValue(50)

    assert len(cl_viewer.axes.images) == 2

    # Return to "No Overlay"
    cubeviz_layout._overlay_controller._overlay_image_combo.setCurrentIndex(0)

    assert len(cl_viewer.axes.images) == 1

    # Remove the moment map data set
    for helper in cubeviz_layout._viewer_combo_helpers:
        helper.remove_data(cubeviz_layout._data.container_2d)
    for cid in cubeviz_layout._data.container_2d.component_ids():
        cubeviz_layout._data.container_2d.remove_component(cid)
    del cubeviz_layout._data.container_2d
    dc = cubeviz_layout.session.data_collection
    dc.remove(dc[1])

def check_viewer_title(cubeviz_layout, viewer_idx, component_idx):
    viewer = cubeviz_layout.cube_views[viewer_idx]._widget

    combo = cubeviz_layout.get_viewer_combo(viewer_idx)
    combo.setCurrentIndex(component_idx)

    component_id = cubeviz_layout.data_components[component_idx]
    component = cubeviz_layout._data.get_component(component_id)

    title = '{} [{}]'.format(component_id, component.units)
    assert viewer.axes.title.get_text() == title


def test_viewer_title(cubeviz_layout):
    """
    Test whether viewer titles accurately reflect data component and flux unit
    """
    for idx in range(len(cubeviz_layout.cube_views[1:])):
        combo = cubeviz_layout.get_viewer_combo(idx)
        current_idx = combo.currentIndex()

        check_viewer_title(cubeviz_layout, idx, 0)
        check_viewer_title(cubeviz_layout, idx, 1)

        # Reset the combo when we're done
        combo = cubeviz_layout.get_viewer_combo(idx)
        combo.setCurrentIndex(current_idx)


def test_viewer_flux_units_change(cubeviz_layout):
    """
    Test whether viewer titles update appropriately with flux unit changes
    """
    # Test after flux units change
    specviz = cubeviz_layout.specviz._widget
    current_units = specviz.hub.plot_widget.data_unit
    specviz.hub.plot_widget.set_data_unit('mJy')

    for idx in range(len(cubeviz_layout.cube_views)):
        combo = cubeviz_layout.get_viewer_combo(idx)
        current_idx = combo.currentIndex()

        check_viewer_title(cubeviz_layout, idx, 0)

        # Check whether the flux units associated with this viewer is correct.
        # This is reflected in the mouseover display of flux values
        assert cubeviz_layout.cube_views[idx]._widget.cubeviz_unit.unit == 'mJy'

        # Reset the combo when we're done
        combo = cubeviz_layout.get_viewer_combo(idx)
        combo.setCurrentIndex(current_idx)

    # Restore original units
    specviz.hub.plot_widget.set_data_unit(current_units)


def test_viewer_wavelength_units_change(cubeviz_layout):
    """
    Test whether image viewer wavelength units are updated appropriately

    These units are used by the viewer to populate the viewer status bar during
    mouseover.
    """

    current_units = str(cubeviz_layout._wavelength_controller.current_units)

    for viewer in cubeviz_layout.cube_views:
        assert viewer._widget._wavelength_units == current_units

    specviz = cubeviz_layout.specviz._widget
    specviz.hub.plot_widget.set_spectral_axis_unit('Angstrom')

    for viewer in cubeviz_layout.cube_views:
        assert viewer._widget._wavelength_units == 'Angstrom'

    # Restore original wavelength units
    specviz.hub.plot_widget.set_spectral_axis_unit(current_units)


def test_default_contour(cubeviz_layout):
    """
    Make sure that default contour works
    """
    # Keep track of children in viewer to check that number increases later
    cl_viewer = cubeviz_layout.split_views[1]._widget
    cl_viewer_children = len(cl_viewer.axes.get_children())

    # Create a subset (ROI) if there is none
    cl_viewer.apply_roi(roi.CircularROI(xc=6, yc=10, radius=3))
    cl_viewer.default_contour()

    assert len(cl_viewer.axes.get_children()) > cl_viewer_children
    assert cl_viewer.is_contour_active

    cl_viewer.remove_contour()

    # Remove the ROI/subset
    dc = cubeviz_layout.session.application.data_collection
    dc.remove_subset_group(dc.subset_groups[0])

    assert cl_viewer.is_contour_active == False
    assert len(cl_viewer.axes.get_children()) == cl_viewer_children

def test_contour_preview(cubeviz_layout):
    """
    Create contour preview from code
    """
    # Keep track of children in viewer to check that number increases later
    cl_viewer = cubeviz_layout.split_views[1]._widget
    cl_viewer_children = len(cl_viewer.axes.get_children())

    preview_settings = ContourSettings(cl_viewer)
    preview_settings.default_options()

    cl_viewer.set_contour_preview(preview_settings)

    assert cl_viewer.is_contour_preview_active
    assert cl_viewer.contour_preview_settings == preview_settings
    assert len(cl_viewer.axes.get_children()) > cl_viewer_children

    cl_viewer.end_contour_preview()

    assert cl_viewer.is_contour_preview_active == False
    assert cl_viewer.contour_preview_settings is None
    assert len(cl_viewer.axes.get_children()) == cl_viewer_children

def test_change_contour_settings(cubeviz_layout):
    """
    Create default contour, open settings menu, change settings, press preview, press ok
    """
    # Keep track of children in viewer to check that number increases later
    cl_viewer = cubeviz_layout.split_views[1]._widget
    cl_viewer_children = len(cl_viewer.axes.get_children())

    cl_viewer.default_contour()
    assert cl_viewer.is_contour_active
    assert len(cl_viewer.axes.get_children()) > cl_viewer_children

    contour_dialog = cl_viewer.edit_contour_settings()

    # Set settings in dialog box
    contour_dialog.contour_label_checkBox.setChecked(True)
    contour_dialog.font_size_input.setText("12")

    contour_dialog.custom_spacing_checkBox.setChecked(True)
    contour_dialog.spacing_input.setText("1")

    contour_dialog.vmax_checkBox.setChecked(True)
    contour_dialog.vmax_input.setText("2")

    contour_dialog.vmin_checkBox.setChecked(True)
    contour_dialog.vmin_input.setText("1")

    contour_dialog.previewButton.click()
    assert cl_viewer.is_contour_preview_active

    contour_dialog.okButton.click()

    # Check to make sure setting changes went through
    assert contour_dialog.font_size_input.text() == "12"
    assert contour_dialog.spacing_input.text() == "1"
    assert contour_dialog.vmax_input.text() == "2"
    assert contour_dialog.vmin_input.text() == "1"

    assert cl_viewer.is_contour_preview_active == False

    cl_viewer.remove_contour()

    assert cl_viewer.is_contour_active == False
    assert len(cl_viewer.axes.get_children()) == cl_viewer_children



import os
import pytest
from glue.core.data_factories import load_data, find_factory
from glue.config import data_factory
from ...listener import CUBEVIZ_LAYOUT

DATA = os.path.join(os.path.dirname(__file__), 'data')


def factory_label(factory):
    # find_factory in glue just returns the factory function, so we have to
    # look at all data factories to figure out which one it is, to make sure
    # the label matches.
    for df in data_factory.members:
        if df.function is factory:
            return df.label


# IMPORTANT: do not include full resolution cubes in the data/ directory.
# Instead, downsample or replace the arrays with smaller arrays.

TEST_CASES = [('manga-7495-12704-LOGCUBE.fits', 'manga', (10, 10, 10)),
              ('SINFONI_CAS64_k_bjames.fits', 'sinfoni-cube', (10, 10, 10)),
              ('det_image_seq1_MIRIFUSHORT_12SHORTexp1_s3d.fits', 'jwst-fits-cube', (10, 10, 10))]


@pytest.mark.parametrize(('filename', 'factory_name', 'shape'), TEST_CASES)
def test_basic(filename, factory_name, shape):

    # This is a basic test to excercise the data factory infrastructure. Above
    # is a list containing (filename, factory, shape) where filename is
    # the name of a file inside the data/ directory, factory is the factory that
    # should be identified to be used for the file, and shape is the expected
    # shape of the data.

    filename = os.path.join(DATA, filename)
    factory = find_factory(filename)
    assert factory_label(factory) == factory_name
    data = load_data(filename)
    assert data.meta[CUBEVIZ_LAYOUT] == factory_name
    assert data.shape == shape



def test_update_specviz(cubeviz_layout):
    """
    Make sure wavelength unit update is reflected in specviz.
    """
    specviz = cubeviz_layout.specviz._widget
    current_units = str(cubeviz_layout._wavelength_controller.current_units)

    cubeviz_layout._wavelength_controller.update_units('Angstrom')
    assert specviz.hub.plot_widget.spectral_axis_unit == 'Angstrom'

    cubeviz_layout._wavelength_controller.update_units(current_units)
    assert specviz.hub.plot_widget.spectral_axis_unit == current_units



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import sys
import os
import argparse

from qtpy.QtCore import QTimer
from qtpy import QtGui, QtWidgets

import glue
from glue.utils.qt import get_qapp
from glue.app.qt import GlueApplication
from glue.main import get_splash, load_data_files, load_plugins

try:
    from glue.utils.qt.decorators import die_on_error
except ImportError:
    from glue.utils.decorators import die_on_error

from . import __version__ as cubeviz_version
from .data_factories import DataFactoryConfiguration


CUBEVIZ_ICON_PATH = os.path.abspath(
    os.path.join(
        os.path.abspath(__file__),
        "..",
        "data",
        "resources",
        "cubeviz_icon.png"
    )
)

CUBEVIZ_LOGO_PATH = os.path.abspath(
    os.path.join(
        os.path.abspath(__file__),
        "..",
        "data",
        "resources",
        "cubeviz_logo.png"
    )
)

def setup():
    from . import layout  # noqa
    from . import startup  # noqa



def _check_datafiles_exist(datafiles):
    for fileparam in datafiles:
        for filename in fileparam.split(','):
            if not os.path.isfile(filename.strip()):
                raise IOError('The file {} does not exist'.format(filename))


def _create_glue_app(data_collection, hub):
    session = glue.core.Session(data_collection=data_collection, hub=hub)
    ga = GlueApplication(session=session)
    ga.setWindowTitle('cubeviz ({})'.format(cubeviz_version))
    qapp = QtWidgets.QApplication.instance()
    qapp.setWindowIcon(QtGui.QIcon(CUBEVIZ_ICON_PATH))
    ga.setWindowIcon(QtGui.QIcon(CUBEVIZ_ICON_PATH))
    return ga


def create_app(datafiles=[], data_configs=[], data_configs_show=False,
               interactive=True):
    """
    Create and initialize a cubeviz application instance

    Parameters
    ----------
    datafiles : `list`
        A list of filenames representing data files to be loaded
    data_configs : `list`
        A list of filenames representing data configuration files to be used
    data_configs_show : `bool`
        Display matching info about data configuration files
    interactive : `bool`
        Flag to indicate whether session is interactive or not (i.e. for testing)
    """
    app = get_qapp()

    # Splash screen
    if interactive:
        splash = get_splash()
        splash.image = QtGui.QPixmap(CUBEVIZ_LOGO_PATH)
        splash.show()
    else:
        splash = None

    # Start off by loading plugins. We need to do this before restoring
    # the session or loading the configuration since these may use existing
    # plugins.
    load_plugins(splash=splash)

    dfc_kwargs = dict(remove_defaults=True, check_ifu_valid=interactive)
    DataFactoryConfiguration(data_configs, data_configs_show, **dfc_kwargs)

    # Check to make sure each file exists and raise an Exception
    # that will show in the popup if it does not exist.
    _check_datafiles_exist(datafiles)

    # Show the splash screen for 1 second
    if interactive:
        timer = QTimer()
        timer.setInterval(1000)
        timer.setSingleShot(True)
        timer.timeout.connect(splash.close)
        timer.start()

    data_collection = glue.core.DataCollection()
    hub = data_collection.hub

    ga = _create_glue_app(data_collection, hub)
    ga.run_startup_action('cubeviz')

    # Load the data files.
    if datafiles:
        datasets = load_data_files(datafiles)
        ga.add_datasets(data_collection, datasets, auto_merge=False)

    if interactive:
        splash.set_progress(100)

    return ga


@die_on_error("Error starting up Cubeviz")
def main(argv=sys.argv):
    """
    The majority of the code in this function was taken from start_glue() in
    main.py. We wanted the ability to get command line arguments and use them
    in here and this seemed to be the cleanest way to do it.

    :param argv:
    :return:
    """

    # # Parse the arguments, ignore any unkonwn
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-configs",
        help="Directory or file for data configuration YAML files",
        action='append', default=[])
    parser.add_argument(
        "--data-configs-show", help="Show the matching info",
        action="store_true", default=False)
    parser.add_argument('data_files', nargs=argparse.REMAINDER)
    args = parser.parse_known_args(argv[1:])

    datafiles = args[0].data_files
    # Store the args for each ' --data-configs' found on the commandline
    data_configs = args[0].data_configs
    data_configs_show = args[0].data_configs_show

    app = create_app(datafiles, data_configs, data_configs_show)
    app.start(maximized=True)



import os
import yaml

from astropy import units as u

DEFAULT_FLUX_UNITS_CONFIGS = os.path.join(os.path.dirname(__file__), 'registered_flux_units.yaml')


def _is_duplicate(unit_list, current_unit):
    """
    Given a list of units, add target units
    after checking for duplicates
    :param unit_list: list of units or string
    :param target_unit: unit or string
    :return: updated unit
    """

    # Make both a string and astropy version
    if isinstance(current_unit, str):
        current_unit = u.Unit(current_unit)

    # Create list of all astropy Unit things
    temp_list = [unit if not isinstance(unit, str) else u.Unit(unit) for unit in unit_list]

    # check to see if current_unit is in there
    return any([unit.to_string() == current_unit.to_string() for unit in temp_list])


class FluxUnitRegistry:
    """
    Saves a list of spectral flux density units.
    """
    def __init__(self):
        self._model_unit = u.Jy  # Will be used for compatibility checks
        self.runtime_defined_units = []  # Stores list of runtime units

    @staticmethod
    def _locally_defined_units():
        """
        list of defined spectral flux density units.
        :return: list of unit strings
        """
        units = ['Jy', 'mJy', 'uJy',
                 'W / (m2 Hz)',
                 'eV / (s m2 Hz)',
                 'erg / (s cm2)',
                 'erg / (s cm2 um)',
                 'erg / (s cm2 Angstrom)',
                 'erg / (s cm2 Hz)',
                 'ph / (s cm2 um)',
                 'ph / (s cm2 Angstrom)',
                 'ph / (s cm2 Hz)']
        return units

    def is_compatible(self, unit):
        """
        Check if unit can be converted to units
        in this registry. Uses model unit to check.
        :param unit: Unit or unit str
        :return: bool
        """
        try:
            self._model_unit.to(unit, equivalencies=u.spectral_density(3500 * u.AA))
            return True
        except u.UnitConversionError:
            return False

    def compose_unit_list(self, current_unit=None):
        """
        Returns a list of unit strings in the registry.
        Adds current_unit if not duplicated. Registered
        units are added first.
        :param current_unit: Unit or unit str
        :return: list of unit str
        """
        unit_list = self._locally_defined_units()  # final list

        for unit in self.runtime_defined_units:
            if not _is_duplicate(unit_list, unit):
                unit_list.append(unit)

        if current_unit is not None:
            if not _is_duplicate(unit_list, current_unit):
                unit_list.append(current_unit)
        return unit_list

    def add_unit(self, item):
        """
        Add new unit
        :param item: unit or unit str
        """
        if isinstance(item, str) \
                or isinstance(item, u.UnitBase):
            if item not in self.runtime_defined_units:
                self.runtime_defined_units.append(item)
        else:
            raise TypeError("Expected unit or string, got {} instead".format(type(item)))


class AreaUnitRegistry:
    """
    Saves a list of area units.
    There are 2 kinds of area units:
    1. solid angles
    2. pixels
    These maybe interchangeable if wcs pixel_scale
    is available.
    """
    def __init__(self):
        self._model_unit = [u.pixel, u.steradian]  # Will be used for compatibility checks
        self.runtime_solid_angle_units = []  # Stores list of runtime angle unitsa
        self.runtime_pixel_units = []  # Stores list of runtime pixel units

    @staticmethod
    def _locally_defined_solid_angle_units():
        """
        list of defined spectral solid angle units.
        :return: list of unit strings
        """
        units = ["deg2",
                 "arcmin2",
                 "arcsec2",
                 'steradian',
                 "rad2"]

        return units

    @staticmethod
    def _locally_defined_pixel_units():
        """
        list of defined spectral pixel units.
        :return: list of unit strings
        """
        units = ["pixel"]

        spaxel = u.def_unit('spaxel', u.astrophys.pix)
        u.add_enabled_units(spaxel)
        units.append('spaxel')

        return units

    def is_compatible(self, unit):
        """
        Check if unit can be converted to units
        in this registry. Uses model units to check.
        :param unit: Unit or unit str
        :return: bool
        """
        compatible = False
        for model_unit in self._model_unit:
            try:
                model_unit.to(unit)
                compatible = True
            except u.UnitConversionError:
                continue
        return compatible

    def compose_unit_list(self, pixel_only=False,
                      solid_angle_only=False,
                      current_unit=None):
        """
        Returns a list of unit strings in the registry.
        Adds current_unit if not duplicated. Registered
        units are added first.
        :param pixel_only: (bool) Return pixel units only
        :param solid_angle_only: (bool) Return solid_angle_only units only
        :param current_unit: Unit or unit str
        :return: list of unit str
        """
        unit_list = []  # final list
        item_list = []  # runtime def units
        if not solid_angle_only:
            unit_list.extend(self._locally_defined_pixel_units())
            item_list.extend(self.runtime_pixel_units)
        if not pixel_only:
            unit_list.extend(self._locally_defined_solid_angle_units())
            item_list.extend(self.runtime_solid_angle_units)

        for unit in item_list:
            if not _is_duplicate(unit_list, unit):
                unit_list.append(unit)

        if current_unit is not None:
            if not _is_duplicate(unit_list, current_unit):
                unit_list.append(current_unit)
        return unit_list

    def add_pixel_unit(self, item):
        """
        Add new pixel unit
        :param item: unit or unit str
        """
        if isinstance(item, str) \
                or isinstance(item, u.UnitBase):
            if item not in self.runtime_pixel_units:
                self.runtime_pixel_units.append(item)
        else:
            raise TypeError("Expected unit or string, got {} instead".format(type(item)))

    def add_solid_angle_unit(self, item):
        """
        Add new solid angle unit
        :param item: unit or unit str
        """
        if isinstance(item, str) \
                or isinstance(item, u.UnitBase):
            if item not in self.runtime_solid_angle_units:
                self.runtime_solid_angle_units.append(item)
        else:
            raise TypeError("Expected unit or string, got {} instead".format(type(item)))

    def add_unit(self, item):
        """
        Check for type and add new unit
        :param item: unit or unit str
        """
        if isinstance(item, str):
            new_unit = u.unit(item)
        else:
            new_unit = item

        if new_unit.decompose() == u.pix.decompose():
            self.add_pixel_unit(new_unit)
        elif 'solid angle' in new_unit.physical_type.lower():
            self.add_solid_angle_unit(new_unit)
        else:
            raise ValueError("Expected pixel or solid angle unit but got {}".format(new_unit.physical_type))


def register_new_unit(new_unit):
    """
    Add new unit to astropy.units
    :param new_unit: astropy unit
    """
    global AREA_UNIT_REGISTRY, FLUX_UNIT_REGISTRY
    u.add_enabled_units(new_unit)
    if FLUX_UNIT_REGISTRY.is_compatible(new_unit):
        FLUX_UNIT_REGISTRY.add_unit(new_unit)
    if new_unit.decompose() == u.pix.decompose() or \
            'solid angle' in new_unit.physical_type.lower():
        AREA_UNIT_REGISTRY.add_unit(new_unit)


def setup_registered_units():
    """
    Function to setup predefined units.
    This should run once when the module
    is imported.
    """
    global FORMATTED_UNITS
    # Load up configurations from yaml file
    with open(DEFAULT_FLUX_UNITS_CONFIGS, 'r') as yamlfile:
        cfg = yaml.load(yamlfile)

    # Define and add new units to astropy
    for new_unit_key in cfg["new_units"]:
        new_unit = cfg["new_units"][new_unit_key]
        try:
            u.Unit(new_unit["name"])
        except ValueError:
            if "base" in new_unit:
                new_astropy_unit = u.def_unit(new_unit["name"], u.Unit(new_unit["base"]))
            else:
                new_astropy_unit = u.def_unit(new_unit["name"])
            register_new_unit(new_astropy_unit)

    new_physical_types = [
        [(u.Jy / u.degree ** 2), 'SFD_over_solid_angle'],
        [(u.Jy / u.pix), 'SFD_over_pix']
    ]

    for model_unit, name in new_physical_types:
        try:
            u.def_physical_type(model_unit, name)
        except ValueError:
            continue

    FORMATTED_UNITS = cfg["formatted_units"]


FLUX_UNIT_REGISTRY = FluxUnitRegistry()
AREA_UNIT_REGISTRY = AreaUnitRegistry()
FORMATTED_UNITS = {}

# Call setup_registered_units on start up:
setup_registered_units()

# CubeViz Unit types:
NONE_CubeVizUnit = "NONE"
UNKNOWN_CubeVizUnit = "UNKNOWN"
ASTROPY_CubeVizUnit = "ASTROPY"
CUBEVIZ_UNIT_TYPES = [NONE_CubeVizUnit, UNKNOWN_CubeVizUnit, ASTROPY_CubeVizUnit]



import pytest
from qtpy import QtCore

import numpy as np
from astropy import units as u

from ...tests.helpers import (enter_slice_text, enter_wavelength_text,
                              left_click, select_viewer, enter_slice_text,
                              toggle_viewer, assert_viewer_indices,
                              assert_all_viewer_indices,
                              assert_wavelength_text, assert_slice_text)

from cubeviz.messages import FluxUnitsUpdateMessage
from ..flux_units import CubeVizUnit, FluxUnitController


def add_get_remove_units(cubeviz_layout, flux_unit_controller=None):
    data = cubeviz_layout._data
    if flux_unit_controller is None:
        flux_unit_controller = FluxUnitController(cubeviz_layout)
        flux_unit_controller.set_data(data)

    component_ids = data.component_ids()
    for component_id in component_ids:
        flux_unit_controller[component_id] = u.Jy

    for component_id in component_ids:
        del flux_unit_controller[component_id]

    for component_id in component_ids:
        flux_unit_controller.add_component_unit(component_id, u.Jy)

    for component_id in component_ids:
        unit = flux_unit_controller.get_component_unit(component_id)
        assert (unit == u.Jy)
        unit = flux_unit_controller.get_component_unit(component_id, cubeviz_unit=True)
        assert (unit.unit == u.Jy)
        unit = flux_unit_controller[component_id]
        assert (unit.unit == u.Jy)

    for component_id in component_ids:
        flux_unit_controller.remove_component_unit(component_id)

    assert (len(flux_unit_controller.components) == 0)


def test_flux_unit_controller(qtbot, cubeviz_layout):
    # test flux controller creation
    flux_unit_controller = FluxUnitController(cubeviz_layout)

    # Test conversion
    cvu = CubeVizUnit(u.Unit("1e-17 erg/s/cm2/Angstrom"),
                      "1e-17 erg/s/cm2/Angstrom",
                      "FLUX",
                      "ASTROPY")

    cvu.controller = flux_unit_controller

    # Set units to something else
    cvu.unit = u.uJy

    wave = 5000 * u.Angstrom
    result = cvu.convert_value(5.0, wave=wave)

    assert(np.allclose(41.69551, result))

    # Test adding, getting and removing units
    # 1) To empty flux unit controller
    add_get_remove_units(cubeviz_layout)

    # 2) To the controller in the layout
    # This has been disabled because it modifies application state in a way
    # that affects other tests. Eventually it may be useful to have a
    # workaround
    #add_get_remove_units(cubeviz_layout, cubeviz_layout._flux_unit_controller)


def test_change_flux_units_specviz(cubeviz_layout):
    """
    Make sure that updates to flux units in cubeviz propagate to specviz.
    """
    specviz = cubeviz_layout.specviz._widget
    flux_controller = cubeviz_layout._flux_unit_controller
    hub = cubeviz_layout.session.hub

    component_id = cubeviz_layout.data_components[0]
    component = cubeviz_layout._data.get_component(component_id)
    current_units = component.units

    cvu = flux_controller.add_component_unit(component_id, 'mJy')
    flux_controller.data.get_component(component_id).units = 'mJy'
    msg = FluxUnitsUpdateMessage(flux_controller, cvu, component_id)
    hub.broadcast(msg)
    assert specviz.hub.plot_widget.data_unit == 'mJy'

    cvu = flux_controller.add_component_unit(component_id, current_units)
    flux_controller.data.get_component(component_id).units = current_units
    msg = FluxUnitsUpdateMessage(flux_controller, cvu, component_id)
    hub.broadcast(msg)
    assert specviz.hub.plot_widget.data_unit == current_units



"""
Contains tests of integration between specviz and cubeviz

In general, most of these tests should be in the specviz->cubeviz direction. In
other words, it should test whether events in specviz cause the appropriate
response in cubeviz. Tests in the other direction are probably better off being
added to existing cubeviz tests.
"""

import pytest
from glue.core import roi


@pytest.mark.parametrize('slice_index', [0, 100, 1000, 1024])
def test_specviz_slice_update(cubeviz_layout, slice_index):
    """
    Make sure slice updates in specviz propagate to cubeviz.
    """
    specviz = cubeviz_layout.specviz._widget
    wavelength = cubeviz_layout._slice_controller._wavelengths[slice_index]

    specviz._slice_indicator.setPos([wavelength, 0])
    assert cubeviz_layout._slice_controller.synced_index == slice_index


def test_cubeviz_roi(cubeviz_layout):
    """
    Test whether ROI creation in cubeviz propagates to specviz
    """
    specviz = cubeviz_layout.specviz._widget
    assert len(specviz._layer_artist_container.layers) == 1

    # Create ROI/subset
    cl_viewer = cubeviz_layout.split_views[1]._widget
    cl_viewer.apply_roi(roi.CircularROI(xc=6, yc=10, radius=3))

    assert len(specviz._layer_artist_container.layers) == 2

    # Remove the ROI/subset
    dc = cubeviz_layout.session.application.data_collection
    dc.remove_subset_group(dc.subset_groups[0])

    assert len(specviz._layer_artist_container.layers) == 1


def test_wavelength_units_update(cubeviz_layout):
    """
    Make sure wavelength unit changes in specviz are reflected in cubeviz.
    """
    specviz = cubeviz_layout.specviz._widget
    current_units = str(cubeviz_layout._wavelength_controller.current_units)

    specviz.hub.plot_widget.set_spectral_axis_unit('Angstrom')
    assert cubeviz_layout._wavelength_controller.current_units == 'Angstrom'

    specviz.hub.plot_widget.set_spectral_axis_unit(current_units)
    assert cubeviz_layout._wavelength_controller.current_units == current_units


def test_flux_units_update(cubeviz_layout):
    """
    Make sure flux unit changes in specviz are reflected in cubeviz
    """
    specviz = cubeviz_layout.specviz._widget

    component_id = cubeviz_layout.data_components[0]
    component = cubeviz_layout._data.get_component(component_id)
    current_units = component.units

    specviz.hub.plot_widget.set_data_unit('mJy')
    assert component.units == 'mJy'

    specviz.hub.plot_widget.set_data_unit(current_units)
    assert component.units == current_units
